/**
 * 	  Copyright (c) 2012 Max Hacking
 *
 *    Authors:  Max Hacking 		<max@hacking.co.uk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "mod_subreqonly_private.h"
#include <apr_dso.h>
#include <ctype.h>

/* Entry point for filter initialisation */
static apr_status_t filter_init(ap_filter_t * f)
{
	// We don't need to do any initialisation but if we did we'd need to call this.
    // svr_cfg *sconf = ap_get_module_config(f->r->server->module_config, &subreqonly_module);

    return APR_SUCCESS;
}

/* Function to determine if this is an ApacheFS sub-request */
static unsigned char is_apachefs_request(ap_filter_t *filter)
{
	// Loop through the previous requests in the chain.  If one of them is
	// an ApacheFS filter then return true.
	ap_filter_t *temp_filter = filter->next;
	while (temp_filter)
	{
		if (strcasecmp(temp_filter->frec->name, APACHEFS_FILTER_NAME) == 0)
			return TRUE;

		temp_filter = temp_filter->next;
	};

	// We got to the end of the filter chain without finding an ApacheFS filter
	// so return false.
    return FALSE;
}

/* Entry point for filter execution */
static apr_status_t filter_func(ap_filter_t * f, apr_bucket_brigade * bb)
{
	dir_cfg *dconf = ap_get_module_config(f->r->per_dir_config, &subreqonly_module);

	// If ALLOW_ALL is set and this is a sub-request of any kind OR
	// ALLOW_APACHE_FS is set and this is an ApacheFS sub-request then
	// get out the way (allowing the request).
	if ( ((dconf->opts & ALLOW_ALL) && !ap_is_initial_req(f->r))	||
		 ((dconf->opts & ALLOW_APACHE_FS) && is_apachefs_request(f)) )
    {
    	ap_remove_output_filter(f);
    	return ap_pass_brigade(f->next, bb);
    }

	// If we got this far then the request was disallowed.
    return HTTP_NOT_FOUND;
}

/* Function to merge a directory configuration */
static void *merge_dir_config(apr_pool_t * p, void *basev, void *addv)
{
    dir_cfg *from = basev;
    dir_cfg *merge = addv;
    dir_cfg *to = apr_palloc(p, sizeof(dir_cfg));

    /* This code comes from mod_autoindex's IndexOptions */
    if (merge->opts & NO_OPTIONS)
    {
        /*
         * If the current directory says 'no options' then we also
         * clear any incremental mods from being inheritable further down.
         */
        to->opts = NO_OPTIONS;
        to->incremented_opts = 0;
        to->decremented_opts = 0;
    }
    else
    {
        /*
         * If there were any nonincremental options selected for
         * this directory, they dominate and we don't inherit *anything.*
         * Contrariwise, we *do* inherit if the only settings here are
         * incremental ones.
         */
        if (merge->opts == 0)
        {
            to->incremented_opts = (from->incremented_opts | merge->incremented_opts) & ~merge->decremented_opts;
            to->decremented_opts = (from->decremented_opts | merge->decremented_opts);
            /*
             * We may have incremental settings, so make sure we don't
             * inadvertently inherit an IndexOptions None from above.
             */
            to->opts = (from->opts & ~NO_OPTIONS);
        }
        else
        {
            /*
             * There are local nonincremental settings, which clear
             * all inheritance from above.  They *are* the new base settings.
             */
            to->opts = merge->opts;;
        }
        /*
         * We're guaranteed that there'll be no overlap between
         * the add-options and the remove-options.
         */
        to->opts |= to->incremented_opts;
        to->opts &= ~to->decremented_opts;
    }

    return to;
}

static void *create_server_cfg(apr_pool_t * p, server_rec * x)
{
    svr_cfg *cfg = apr_pcalloc(p, sizeof(svr_cfg));

    cfg->announce = 1;

    return cfg;
}

static void *create_dir_config(apr_pool_t * p, char *x)
{
    dir_cfg *conf = apr_pcalloc(p, sizeof(dir_cfg));

    conf->opts = 0;
    conf->incremented_opts = 0;
    conf->decremented_opts = 0;

    return conf;
}

static const char *add_opts(cmd_parms * cmd, void *d, const char *optstr)
{
    char *w;
    apr_int32_t opts;
    apr_int32_t opts_add;
    apr_int32_t opts_remove;
    char action;
    dir_cfg *d_cfg = (dir_cfg *) d;

    opts = d_cfg->opts;
    opts_add = d_cfg->incremented_opts;
    opts_remove = d_cfg->decremented_opts;
    while (optstr[0])
    {
        int option = 0;

        w = ap_getword_conf(cmd->pool, &optstr);

        if ((*w == '+') || (*w == '-'))
        {
            action = *(w++);
        }
        else
        {
            action = '\0';
        }


        if (!strcasecmp(w, "AllowApacheFS"))
        {
            option = ALLOW_APACHE_FS;
        }
        else if (!strcasecmp(w, "AllowAll"))
        {
            option = ALLOW_ALL;
        }
        else if (!strcasecmp(w, "None"))
        {
            if (action != '\0')
            {
                return "Cannot combine '+' or '-' with 'None' keyword";
            }
            opts = NO_OPTIONS;
            opts_add = 0;
            opts_remove = 0;
        }
        else
        {
            return "Invalid SubReqOnlyOption";
        }

        if (action == '\0')
        {
            opts |= option;
            opts_add = 0;
            opts_remove = 0;
        }
        else if (action == '+')
        {
            opts_add |= option;
            opts_remove &= ~option;
        }
        else
        {
            opts_remove |= option;
            opts_add &= ~option;
        }
    }
    if ((opts & NO_OPTIONS) && (opts & ~NO_OPTIONS))
    {
        return "Cannot combine other TransformOptions keywords with 'None'";
    }
    d_cfg->incremented_opts = opts_add;
    d_cfg->decremented_opts = opts_remove;
    d_cfg->opts = opts;
    return NULL;
}

/* This function is registered as a command hook for the SubReqOnlyAnnounce command */
static const char *set_announce(cmd_parms *cmd, void *struct_ptr, int arg)
{
    svr_cfg *cfg = ap_get_module_config(cmd->server->module_config,	&subreqonly_module);

    // SubReqOnlyAnnounce is a server wide option, check it has been used as such.
    const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE | NOT_IN_LIMIT);
    if (err)
        return err;

    cfg->announce = arg ? 1 : 0;

    return NULL;
}

/* Once the module is loaded, initialised, the hooks registered and the server wide
 * configuration loaded this function is executed.  Server wide configuration options
 * should be acted upon here if required.
 */
static int post_config_hook(apr_pool_t *p, apr_pool_t *log, apr_pool_t *ptemp, server_rec *s)
{
    svr_cfg *cfg = ap_get_module_config(s->module_config, &subreqonly_module);

    /* Add version string to Apache headers */
    if (cfg->announce)
    {
        char *compinfo = PACKAGE_NAME "/" PACKAGE_VERSION;

        ap_add_version_component(p, compinfo);
    }

    return OK;
}

/* The register_hooks function will be called once when the module is loaded
 * and should only contain calls to hook and filter registration functions.
 */
static void register_hooks(apr_pool_t * p)
{
    ap_hook_post_config(post_config_hook, NULL, NULL, APR_HOOK_MIDDLE);

    ap_register_output_filter(SUBREQONLY_FILTER_NAME, filter_func, filter_init, AP_FTYPE_RESOURCE);
};

/* The command array.  This contains entries for the commands
 * which appear in configuration files.
 */
static const command_rec command_array[] =
{
    AP_INIT_RAW_ARGS("SubReqOnlyOptions", add_opts, NULL, OR_INDEXES, "One or more index options [+|-][]"),

    AP_INIT_FLAG("SubReqOnlyAnnounce", set_announce, NULL, RSRC_CONF, "Whether to announce this module in the server header. Default: On"),

    {NULL}
};

/* The declaration of the Apache module.  This structure contains function pointers to
 * connect the module to the daemon.
 */
module AP_MODULE_DECLARE_DATA subreqonly_module =
{
    STANDARD20_MODULE_STUFF,
    create_dir_config,
    merge_dir_config,
    create_server_cfg,
    NULL,
    command_array,
    register_hooks
};
